package bk;

import java.util.Stack;

import javax.swing.JTree;

import animation.AlgorithmStage;
import animation.BackwardAction;
import animation.PseudoCodeNode;
import animation.PseudoCodeNode.CodeAction;
import graph.LayeredGraphEdge;
import graph.LayeredGraphNode;
import lib.TextLayoutHelper;

/**
 * the preprocessing stage of the bk node placement algorithm
 * 
 * @author kolja and eren
 *
 */
public class ConflictDetection implements AlgorithmStage {

    private LayeredGraphNode graph;
    private Stack<BackwardAction> actions;

    private int i;
    private int l1;
    private int k0;
    private int k1;
    private int l;
    private int k;
    private int hidden_k;
    /**
     * line number in Carstens' pseudocode listing 3.1
     */
    private int pseudo_line;
    private PseudoCodeNode lines[];
    private boolean breakPoint;
    private boolean skip;

    public ConflictDetection(LayeredGraphNode graph) {
        this.graph = graph;
        actions = new Stack<>();
        i = 0; // will be increased before first iteration
        l1 = -1; // will be increased before first iteration
        k0 = 0;
        k1 = 0;
        l = 0;
        k = 0;
        hidden_k = 0;
        pseudo_line = 1;
        breakPoint = false;
        skip = false;
    }


    @Override
    public StageStatus forwardStep() {
        int old_line = pseudo_line;
        int old_k = k;
        int old_k0 = k0;
        int old_k1 = k1;
        int old_l = l;
        int old_l1 = l1;
        int old_hidden_k = hidden_k;

        switch (pseudo_line) {
        case 1:
            i++;
            actions.push(() -> {
                i--;
                pseudo_line = old_line;
                selectNodes(pseudo_line);
            });
            if (i <= graph.getContainedLayers().size() - 2) {
                pseudo_line++;
            } else {
                for( PseudoCodeNode p : lines )
                {
                    if( p != null )
                        p.setSelected( false );
                }
                return StageStatus.FINISHED;
            }
            break;
        case 2:
            k0 = 0;
            l = 0;
            l1 = -1; // will be increased before first iteration
            pseudo_line++;
            actions.push(() -> {
                pseudo_line = old_line;
                k0 = old_k0;
                l = old_l;
                l1 = old_l1;
                selectNodes(pseudo_line);
            });
            break;
        case 3:
            l1++;
            actions.push(() -> {
                l1--;
                pseudo_line = old_line;
                selectNodes(pseudo_line);
            });
            if (l1 < graph.getContainedLayers().get(i + 1).size()) {
                pseudo_line += 1;
            } else {
                pseudo_line = 1;
            }
            break;
        case 4:
            actions.push(() -> {
                pseudo_line = old_line;
                selectNodes(pseudo_line);
            });
            if (graph.getContainedLayers().get(i + 1).size() == l1 || incidentToInnerSegmentBetweenLiPlusOneAndLi()) {
                pseudo_line++;
            } else {
                pseudo_line = 3;
            }
            break;
        case 5:
            actions.push(() -> {
                k1 = old_k1;
                pseudo_line = old_line;
                selectNodes(pseudo_line);
            });
            k1 = graph.getContainedLayers().get(i).size() - 1;
            pseudo_line++;
            break;
        case 6:
            actions.push(() -> {
                pseudo_line = old_line;
                selectNodes(pseudo_line);
            });
            if (incidentToInnerSegmentBetweenLiPlusOneAndLi()) {
                pseudo_line++;
            } else {
                pseudo_line = 9;
            }
            break;
        case 7:
            actions.push(() -> {
                k1 = old_k1;
                pseudo_line = old_line;
                selectNodes(pseudo_line);
            });
            k1 = graph.getContainedLayers().get(i).indexOf(
                    graph.getContainedLayers().get(i + 1).get(l1).getSortedIncomingEdges().get(0).getSources().get(0));
            pseudo_line = 9;
            break;
        case 9:
            actions.push(() -> {
                hidden_k = old_hidden_k;
                pseudo_line = old_line;
                selectNodes(pseudo_line);
            });
            if (l <= l1) {
                pseudo_line++;
                hidden_k = 0; // will not be increased before first iteration
                k = graph.getContainedLayers().get(i).indexOf(graph.getContainedLayers().get(i + 1).get(l)
                        .getSortedIncomingEdges().get(hidden_k).getSources().get(0));
            } else {
                pseudo_line = 15;
            }
            break;
        case 10:
            actions.push(() -> {
                k = old_k;
                pseudo_line = old_line;
                selectNodes(pseudo_line);
            });
            if (hidden_k < graph.getContainedLayers().get(i + 1).get(l).getSortedIncomingEdges().size()) {
                k = graph.getContainedLayers().get(i).indexOf(graph.getContainedLayers().get(i + 1).get(l)
                        .getSortedIncomingEdges().get(hidden_k).getSources().get(0));
                pseudo_line++;
            } else {
                pseudo_line = 13;
            }
            break;
        case 11:
            actions.push(() -> {
                hidden_k--;
                graph.getContainedLayers().get(i + 1).get(l).getSortedIncomingEdges().get(hidden_k).setConflicted(false,
                        null);
                pseudo_line = old_line;
                selectNodes(pseudo_line);
            });
            if (k < k0 || k > k1) {
                graph.getContainedLayers().get(i + 1).get(l).getSortedIncomingEdges().get(hidden_k).setConflicted(true,
                        null);
            }
            hidden_k++;
            pseudo_line = 10;
            break;
        case 13:
            actions.push(() -> {
                l--;
                pseudo_line = old_line;
                selectNodes(pseudo_line);
            });
            l++;
            pseudo_line = 9;
            break;
        case 15:
            actions.push(() -> {
                k0 = old_k0;
                pseudo_line = old_line;
                selectNodes(pseudo_line);
            });
            k0 = k1;
            pseudo_line = 3;
            break;
        }

        selectNodes(pseudo_line);
        if( breakPoint )
            return StageStatus.BREAKPOINT;
        if( skip )
            return forwardStep();
        return StageStatus.UNFINISHED;
    }

    private void selectNodes(int next_line) {
        if (next_line >= 3 && l < graph.getContainedLayers().get(i + 1).size()) {
            graph.getContainedLayers().get(i + 1).get(l).setSelected(null);
        }

        if (next_line >= 4) {
            graph.getContainedLayers().get(i + 1).get(l1).setSelected(null);
        }

        if (next_line >= 3) {
            graph.getContainedLayers().get(i).get(k0).setSelected(null);
        }

        if (next_line >= 6) {
            graph.getContainedLayers().get(i).get(k1).setSelected(null);
        }

        if (next_line == 10 || next_line == 11) {
            graph.getContainedLayers().get(i).get(k).setSelected(null);
        }
        breakPoint = false;
        skip = false;
        for( PseudoCodeNode p : lines )
        {
            if( p != null )
                p.setSelected( false );
        }
        markCode( lines[ next_line - 1 ] );
    }
    
    private int getCodeDepth( PseudoCodeNode n )
    {
        if( n == lines[ 0 ] )
            return 0;
        return 1 + getCodeDepth( (PseudoCodeNode)n.getParent() );
    }
    
    private void markCode( PseudoCodeNode n )
    {
        if( !n.hasSelectedSubnode() )
        {
            CodeAction action = n.setSelected( true );
            breakPoint |= action == CodeAction.STOP;
            skip |= action == CodeAction.SKIP;
        }
        else
        {
            CodeAction action = n.setSelected( true );
            skip |= action == CodeAction.SKIP;
        }
        if( n == lines[ 0 ] )
            return;
        markCode( (PseudoCodeNode)n.getParent() );
    }

    @Override
    public StageStatus backwardStep() {
        if (actions.size() == 0) {
            for( PseudoCodeNode p : lines )
            {
                if( p != null )
                    p.setSelected( false );
            }
            return StageStatus.FINISHED;
        }
        actions.pop().reverse();
        if( breakPoint )
            return StageStatus.BREAKPOINT;
        if( skip )
            return forwardStep();
        return StageStatus.UNFINISHED;
    }

    @Override
    public PseudoCodeNode createPseudocodeTree(JTree tree) {
        PseudoCodeNode root = new PseudoCodeNode("Preprocessing (mark type 1 conflicts)", tree);
        lines = new PseudoCodeNode[15];
        lines[ 0 ] = new PseudoCodeNode( "for i=1 to |L|-2 do", tree );
        lines[ 1 ] = new PseudoCodeNode( "k0 = 0; l = 0;", tree );
        lines[ 0 ].add( lines[ 1 ] );
        lines[ 2 ] = new PseudoCodeNode( "for l1=0 to |L[i+1]|-1 do", tree );
        lines[ 0 ].add( lines[ 2 ] );
        lines[ 3 ] = new PseudoCodeNode( "if l1==|L[i+1]|-1 or L[i+1][l1] incident to inner segment between L[i+1] and L[i] then", tree );
        lines[ 2 ].add( lines[ 3 ] );
        lines[ 4 ] = new PseudoCodeNode( "k1 = |Li|-1;", tree );
        lines[ 3 ].add( lines[ 4 ] );
        lines[ 5 ] = new PseudoCodeNode( "if L[i+1][l1] incident to inner segment between L[i+1] and L[i] then", tree );
        lines[ 3 ].add( lines[ 5 ] );
        lines[ 6 ] = new PseudoCodeNode( "k1 = pos(pred(L[i+1][l1])[0]);", tree );
        lines[ 5 ].add( lines[ 6 ] );
        lines[ 8 ] = new PseudoCodeNode( "while l <= l1 do", tree );
        lines[ 3 ].add( lines[ 8 ] );
        lines[ 9 ] = new PseudoCodeNode( "foreach v in pred(L[i+1][l]) do", tree );
        lines[ 8 ].add( lines[ 9 ] );
        lines[ 10 ] = new PseudoCodeNode( "if k < k0 or k > k1 then mark segment (v,L[i+1][l]);", tree );
        lines[ 9 ].add( lines[ 10 ] );
        lines[ 12 ] = new PseudoCodeNode( "l = l+1;", tree );
        lines[ 8 ].add( lines[ 12 ] );
        lines[ 14 ] = new PseudoCodeNode( "k0 = k1;", tree );
        lines[ 3 ].add( lines[ 14 ] );
        root.add( lines[ 0 ] );
        return root;
    }

    @Override
    public StageStatus forwardStepOver() {
        int depth = getCodeDepth( lines[ pseudo_line - 1 ] );
        do
        {
            switch( forwardStep() )
            {
            case BREAKPOINT:
                return StageStatus.BREAKPOINT;
            case FINISHED:
                return StageStatus.FINISHED;
            case UNFINISHED:
                break;
            }
        } while( depth < getCodeDepth( lines[ pseudo_line - 1 ] ) );
        return StageStatus.UNFINISHED;
    }

    @Override
    public StageStatus forwardStepOut() {
        int depth = getCodeDepth( lines[ pseudo_line - 1 ] );
        do
        {
            switch( forwardStep() )
            {
            case BREAKPOINT:
                return StageStatus.BREAKPOINT;
            case FINISHED:
                return StageStatus.FINISHED;
            case UNFINISHED:
                break;
            }
        } while( depth <= getCodeDepth( lines[ pseudo_line - 1 ] ) );
        return StageStatus.UNFINISHED;
    }

    @Override
    public StageStatus backwardStepOver() {
        int depth = getCodeDepth( lines[ pseudo_line - 1 ] );
        do
        {
            switch( backwardStep() )
            {
            case BREAKPOINT:
                return StageStatus.BREAKPOINT;
            case FINISHED:
                return StageStatus.FINISHED;
            case UNFINISHED:
                break;
            }
        } while( depth < getCodeDepth( lines[ pseudo_line - 1 ] ) );
        return StageStatus.UNFINISHED;
    }

    @Override
    public StageStatus backwardStepOut() {
        int depth = getCodeDepth( lines[ pseudo_line - 1 ] );
        do
        {
            switch( backwardStep() )
            {
            case BREAKPOINT:
                return StageStatus.BREAKPOINT;
            case FINISHED:
                return StageStatus.FINISHED;
            case UNFINISHED:
                break;
            }
        } while( depth <= getCodeDepth( lines[ pseudo_line - 1 ] ) );
        return StageStatus.UNFINISHED;
    }

    @Override
    public String getDebugString() {
        String info = "| i  | l  | l1 | k0 | k1 | k  |\n";
        info +=       "|----|----|----|----|----|----|\n";
        info += "|" + TextLayoutHelper.strToLen( "" + i, 4 ) + 
                "|" + TextLayoutHelper.strToLen( "" + l, 4 ) + 
                "|" + TextLayoutHelper.strToLen( "" + l1, 4 ) + 
                "|" + TextLayoutHelper.strToLen( "" + k0, 4 ) + 
                "|" + TextLayoutHelper.strToLen( "" + k1, 4 ) + 
                "|" + TextLayoutHelper.strToLen( "" + k, 4 ) + "|\n";
//        if( insideSubgraph && vIndex < graph.getContainedNodes().size() )
//        {
//            info += "Subgraph of " + graph.getContainedNodes().get( vIndex ).getName() + ":\n";
//            String tmp = subgraphAlgs.get( vIndex ).getDebugString();
//            info += tmp;
//            return info;
//        }
        return info;
    }

    /**
     * part of line 4 and 6 in the pseudocode
     * 
     * @return
     */
    private boolean incidentToInnerSegmentBetweenLiPlusOneAndLi() {
        LayeredGraphNode curr = graph.getContainedLayers().get(i + 1).get(l1);
        for (LayeredGraphEdge e : curr.getIncomingEdges()) {
            if (e.isDummyEdge()) {
                return true;
            }
        }
        return false;
    }
}
